package build

import (
	"os"
	"path/filepath"
	"strings"
	"sync"

	"git.akyoto.dev/cli/q/src/build/directory"
	"git.akyoto.dev/cli/q/src/build/token"
)

// Scan scans the directory.
func Scan(files []string) (<-chan *Function, <-chan error) {
	functions := make(chan *Function)
	errors := make(chan error)

	go func() {
		scan(files, functions, errors)
		close(functions)
		close(errors)
	}()

	return functions, errors
}

// scan scans the directory without channel allocations.
func scan(files []string, functions chan<- *Function, errors chan<- error) {
	wg := sync.WaitGroup{}

	for _, file := range files {
		stat, err := os.Stat(file)

		if err != nil {
			errors <- err
			return
		}

		if stat.IsDir() {
			err = directory.Walk(file, func(name string) {
				if !strings.HasSuffix(name, ".q") {
					return
				}

				fullPath := filepath.Join(file, name)
				wg.Add(1)

				go func() {
					defer wg.Done()
					err := scanFile(fullPath, functions)

					if err != nil {
						errors <- err
					}
				}()
			})

			if err != nil {
				errors <- err
			}
		} else {
			wg.Add(1)

			go func() {
				defer wg.Done()
				err := scanFile(file, functions)

				if err != nil {
					errors <- err
				}
			}()
		}
	}

	wg.Wait()
}

// scanFile scans a single file.
func scanFile(path string, functions chan<- *Function) error {
	contents, err := os.ReadFile(path)

	if err != nil {
		return err
	}

	tokens := token.Tokenize(contents)

	var (
		groupLevel  = 0
		blockLevel  = 0
		headerStart = -1
		bodyStart   = -1
	)

	for i, t := range tokens {
		switch t.Kind {
		case token.Identifier:
			if blockLevel == 0 && groupLevel == 0 {
				headerStart = i
			}

		case token.GroupStart:
			groupLevel++

		case token.GroupEnd:
			groupLevel--

		case token.BlockStart:
			blockLevel++

			if blockLevel == 1 {
				bodyStart = i
			}

		case token.BlockEnd:
			blockLevel--

			if blockLevel == 0 {
				function := &Function{
					Name: tokens[headerStart].Text(),
					Head: tokens[headerStart:bodyStart],
					Body: tokens[bodyStart : i+1],
				}

				functions <- function
			}
		}
	}

	return nil
}
